'''
Author: Devin
Date: 2024-07-03 16:07:12
LastEditors: Devin
LastEditTime: 2024-10-09 11:08:40
Description: 

Copyright (c) 2024 by Devin, All Rights Reserved. 
'''
import os
import pickle
import netCDF4 as nc
import numpy as np
from tqdm.auto import tqdm
from netCDF4 import Dataset
from wrf import getvar, interplevel, to_np, latlon_coords,ALL_TIMES,extract_times
from wrf.g_rh import get_rh, get_rh_2m
from wrf.g_uvmet import get_uvmet_wspd,get_uvmet10_wspd,get_uvmet_wdir,get_uvmet10_wdir #,get_uvmet_wspd_wdir,get_uvmet10_wspd_wdir,get_uvmet,get_uvmet10
from netcdf_helper import combine_netcdf_files_with_mfdataset
#气象站点的气象变量和wrf输出文件中的变量对应关系
wrf_variables = {"Planetary_Boundary_Layer_Height":["PBLH"],
                 'Temperature_2m': ['T2'],#2m temperature,对应气象站点的温度
                 'Temperature': ['T'],
                 'Humidity': ["T","Q","PSFC"],
                 'Humidity_2m': ['T2',"Q2","PSFC"],#2m relative humidity,对应气象站点的相对湿度
                 'Pressure_Surface': ['PSFC'],
                 'Wind_Speed_10m': ['U10', 'V10'], # 10m wind speed,对应气象站点的风速
                 'Wind_Direction_10m': ['U10', 'V10'],# 10m wind direction,对应气象站点的风向
                 'Wind_Speed': ['U', 'V'], # 风速
                 'Wind_Direction': ['U', 'V'], # 风向
                 'Solar_Radiation': ['SWDOWN'],
                 'Precipitation': ['RAINC', 'RAINNC'],
                 'Cloud_Cover': ['CLDFRA']
                 }

def get_model_lat_lon_proj(model_lat_lon_file,variable_name="T2", model_pkl_file=None):
    dir=os.path.dirname(model_lat_lon_file)
    model_file=os.path.basename(model_lat_lon_file)
    model_pkl_file=os.path.join(dir,f'{model_file}.pkl') if model_pkl_file is None else model_pkl_file
    if os.path.exists(model_pkl_file):
        with open(model_pkl_file, 'rb') as file:
            cart_proj, model_lats, model_lons = pickle.load(file)
    else:       
        x,y,model_lons,model_lats,dx,dy,cart_proj=get_wrf_info(model_lat_lon_file,variable_name="T2")
        # model_lats, model_lons= to_np(lats), to_np(lons)
        with open(model_pkl_file, 'wb') as file:
            pickle.dump((cart_proj, model_lats, model_lons), file)
    return cart_proj,model_lats,model_lons

def get_wrf_info(wrf_out_file,variable_name="T2"):
    '''
    Get the x,y,lons,lats,dx,dy,cart_proj from wrf_out_file
    @param: wrf_out_file: the wrf output file
    @param: variable_name: the variable name to get the info
    @return: x,y,lons,lats,dx,dy,cart_proj: the x,y grid (lamber projection), lons,lats grid, dx,dy(grid size) and cartopy projection

    '''
    from wrf import (to_np, getvar, get_cartopy, cartopy_xlim,
                 cartopy_ylim, latlon_coords)
    import netCDF4 as nc
    dataset = nc.Dataset(wrf_out_file, 'r')
    t2 = getvar(dataset,variable_name)
    cart_proj = get_cartopy(t2)
    lats, lons =latlon_coords(t2)
    lons,lats =to_np(lons), to_np(lats)
    xlims,ylims= cartopy_xlim(t2),cartopy_ylim(t2)
    x_coords,y_coords=np.linspace(xlims[0],xlims[1],lons.shape[1]),np.linspace(ylims[0],ylims[1],lons.shape[0])   
    x,y=np.meshgrid(x_coords,y_coords) 
    dx, dy=dataset.DX,dataset.DY 
    return x,y,lons,lats,dx,dy,cart_proj 

def extract_data_from_wrf_file(wrf_file,variables=["Planetary_Boundary_Layer_Height", "Temperature_2m", "Humidity_2m", "Pressure_Surface", "Wind_Speed_10m", "Wind_Direction_10m","Solar_Radiation","Precipitation","Cloud_Cover"],date_peroid=None,convert_to_beijing_time=False):
    '''
    Extract data from WRF output file
    @param {str}: wrf_file: the WRF output file
    @param {list}: variables: the variables to extract.see wrf_variables for details。Should be one or more of wrf_variables.keys()
    @param {pd.date_range}: date_peroid: the date period to extract data；示例 date_peroid=pd.date_range("2022-09-01 00:00:00", "2022-10-01 00:00:00", freq="H")或date_peroid=[start_time,end_time], like [start_time,end_time].If None, extract all data.
    @return: 
    {dict}: variables_data: the extracted variables data
    {dict}: variables_units: the units of extracted variables    
    {dict}: global_attrs_template: the global attributes of WRF output file
    {dict}: dict_dims: the dimensions of extracted variables
    {np.array}: lats: the latitudes of WRF output file
    {np.array}: lons: the longitudes of WRF output file
    {np.array}: times: the times of WRF output file
    示例：
    dict_variable_data,dict_variable_units,global_attrs_template,dict_dims,lats,lons,time=extract_wrf_data(wrf_file,variables=["Temperature_2m", "Humidity_2m", "Pressure_Surface", "Wind_Speed_10m", "Wind_Direction_10m"])
    '''
    # 打开 WRF 文件
    ds_input = Dataset(wrf_file) 
    dict_variable_data,dict_variable_units,dict_dims = {},{},{}         
    for variable in variables:                      
        if variable in ["Wind_Speed_10m","Wind_Direction_10m","Wind_Speed","Wind_Direction"]:
            if variable =="Wind_Speed_10m":
                var_obj=get_uvmet10_wspd(ds_input, timeidx=ALL_TIMES)
            if variable =="Wind_Direction_10m":
                var_obj=get_uvmet10_wdir(ds_input, timeidx=ALL_TIMES)
            if variable =="Wind_Speed":
                var_obj=get_uvmet_wspd(ds_input, timeidx=ALL_TIMES)
            if variable =="Wind_Direction":
                var_obj=get_uvmet_wdir(ds_input, timeidx=ALL_TIMES)                              
            # # 计算风速和风向       
            # wind_speed10, wind_direction10 = get_uvmet10_wspd_wdir(ds_input, timeidx=ALL_TIMES)               
        elif variable in ["Humidity_2m","Humidity"]:
            # 计算相对湿度
            var_obj=get_rh_2m(ds_input, timeidx=ALL_TIMES) if variable=="Humidity_2m" else get_rh(ds_input, timeidx=ALL_TIMES)              
        elif variable =="Precipitation":
            # 计算降水量
            rainc=getvar(ds_input, "RAINC",timeidx=ALL_TIMES)  # RAINC：累计对流降水量
            rainnc=getvar(ds_input, "RAINNC",timeidx=ALL_TIMES)  # RAINNC：累计非对流降水量。        
            var_obj = rainc + rainnc  # 总降水量             
        else:
            wrf_var=wrf_variables[variable][0]
            var_obj = getvar(ds_input, wrf_var,timeidx=ALL_TIMES)
        if date_peroid is not None:
            # times = extract_times(ds_input, timeidx=ALL_TIMES)  # 返回 datetime 对象列表
            # # 筛选指定时间段的数据
            if convert_to_beijing_time:
                hours_offset=8
            else:
                hours_offset=0
            start_time,end_time=date_peroid[0]-pd.Timedelta(hours=hours_offset),date_peroid[-1]-pd.Timedelta(hours=hours_offset)
            # time_mask = [(t >= start_time) & (t <= end_time) for t in times] 
            # filtered_var_obj1 = var_obj[time_mask,...]  # 根据时间掩码筛选数据            
            var_obj = var_obj.sel(Time=slice(start_time, end_time))
            # dict_variable_data[variable] = to_np(filtered_var_obj) if variable not in ["Temperature_2m","Temperature"] else to_np(filtered_var_obj)- 273.15 # 转换为摄氏度
            # dict_variable_units[variable] = (filtered_var_obj.units if variable !="Precipitation" else rainc.units) if variable not in ["Temperature_2m","Temperature"] else "°C"
            # dict_dims[variable] = filtered_var_obj.dims    
        # else:             
        dict_variable_data[variable] = to_np(var_obj) if variable not in ["Temperature_2m","Temperature"] else to_np(var_obj)- 273.15 # 转换为摄氏度       
        dict_variable_units[variable] = (var_obj.units if variable !="Precipitation" else rainc.units) if variable not in ["Temperature_2m","Temperature"] else "°C"
        dict_dims[variable] = var_obj.dims            
    # 读取所有全局属性
    global_attrs_template = {
        attr: getattr(ds_input, attr) for attr in ds_input.ncattrs()
    }     
    # 提取时间和经纬度
    # time = getvar(ds_input, "Times").values  # 时间,只获取第一个时间
    times = var_obj.Time.data#extract_times(ds_input, timeidx=ALL_TIMES)  # 返回 datetime 对象列表
    lats, lons = latlon_coords(var_obj)  # 经纬度
    return dict_variable_data,dict_variable_units,dict_dims,global_attrs_template,lats,lons,times
 
def save_wrf_data(output_file,dict_dims,global_attrs_template,variables_data,variables_units=None,output_file_format="NETCDF3_CLASSIC"):
    '''
    @description: Save extracted data to a new WRF output file.
    @param {str}: output_file: the output file name
    @param {dict}: dict_dims: the dimensions of extracted variables
    @param {dict}: global_attrs_template: the global attributes of WRF output file
    @param {dict}: variables_data: the extracted variables data
    @param {dict}: variables_units: the units of extracted variables
    @param {str}: output_file_format: the output file format, default is NETCDF3_CLASSIC
    @return: None
    '''
    output_path = os.path.dirname(output_file)
    output_file_name = os.path.basename(output_file)
    if not os.path.exists(output_path):
        os.makedirs(output_path)    
    ds_output = nc.Dataset(
        output_file, "w", format=output_file_format
    )
    # 写入全局属性
    for attr_name, attr_value in global_attrs_template.items():    
        setattr(ds_output, attr_name, attr_value)
    # 创建变量
    for var_name, values in tqdm(variables_data.items(),leave=True,desc=f"Creating variables for {output_file_name}",):       
        # 写入变量属性
        dims=dict_dims[var_name]
        for idx,dim in enumerate(dims):
            if dim not in ds_output.dimensions:
                ds_output.createDimension(dim,values.shape[idx]) 
        # 创建变量       
        nc_var = ds_output.createVariable(var_name, np.float32, dims)      
        try:
            nc_var.units = variables_units[var_name] if variables_units is not None else ""
            nc_var.long_name = var_name             
            nc_var[:] = values
            ds_output.sync()  # 手动同步数据到磁盘
        except Exception as e:
            print(f"Error setting attributes for {var_name}: {e}")
    ds_output.close()
   
def extract_wrf_hourly_data(wrf_out_dir, output_wrf_file,variables=None,wrf_out_prefix="wrfout_d03_",output_wrf_file_format="NETCDF3_CLASSIC",date_peroid=None,convert_to_beijing_time=False,support_wrf_python=True):
    '''
    @description: 从 WRF 输出文件中提取指定date_peroid,变量variables的数据
    @param {str}: wrf_out_dir: WRF 输出文件所在目录
    @param {str}: output_wrf_file: 输出文件名
    @param {list}: variables: 需要提取的变量列表,如果为None，则提取所有变量,即仅仅提取文件。否则，提取指定变量,指定变量需在 wrf_variables 中定义
    @param {str}: wrf_out_prefix: WRF 输出文件名前缀，默认为 wrfout_d03_
    @param {str}: output_wrf_file_format: 输出文件格式，默认为 NETCDF3_CLASSIC
    @param {pd.date_range}: date_peroid: 日期范围，如果为None，则提取所有时间段的数据。否则，提取指定时间段的数据
    @param {bool}: convert_to_beijing_time: 是否将时间转换为北京时间，默认不转换
    @param {bool}: support_wrf_python: 是否支持 wrf-python 库，默认支持，会额外提取 XLAT , XLONG 和 Times 变量；如果不支持，则只提取variables中定义的变量
    @return: {dict}: dict_variable_data: 变量数据字典，key 为变量名,保持和variables中的顺序一致，value 为变量数据
    @return: {dict}: dict_variable_units: 变量单位字典,key 为变量名，value 为变量单位
    @return: {dict}: dict_dims: 变量维度字典,key 为变量名，value 为变量维度
    @return: {dict}: global_attrs_template: 全局属性字典
    示例：
    variables_data,variables_units,start_time=extract_wrf_hourly_data(wrf_out_dir, output_wrf_file,variables=variables,wrf_out_prefix="wrfout_d03_",output_wrf_file_format="NETCDF3_CLASSIC",date_peroid=date_peroid,convert_to_beijing_time=True)
    '''
    wrf_files = sorted(glob.glob(os.path.join(wrf_out_dir, f"{wrf_out_prefix}*")))
    extracted_wrf_variables =list(set([var for variable in variables for var in wrf_variables[variable]]) if variables is not None else [])
    combine_netcdf_files_with_mfdataset(wrf_files, output_wrf_file, extract_wrf_variables=extracted_wrf_variables, output_file_format=output_wrf_file_format, support_wrf_python=support_wrf_python)
    dict_variable_data,dict_variable_units,dict_dims,global_attrs_template,lats,lons,times=extract_data_from_wrf_file(output_wrf_file,variables=variables,date_peroid=date_peroid,convert_to_beijing_time=convert_to_beijing_time)
    return dict_variable_data,dict_variable_units,times

def extract_wrf_hourly_data_without_combining_files(wrf_out_dir, output_wrf_file,variables,wrf_out_prefix="wrfout_d03_",output_wrf_file_format="NETCDF3_CLASSIC",date_peroid=None,convert_to_beijing_time=False,support_wrf_python=True):
    '''
    Extract hourly data from WRF output files and save them to a new WRF output file.
    @param {str}: wrfout_dir: the directory of WRF output files
    @param {str}: output_wrf_file: extracted WRF output file
    @param {list}: variables: the variables to extract.see wrf_variables for details。Should be one or more of wrf_variables.keys()
    @param {str}: output_wrf_file_format: the output file format, default is NETCDF3_CLASSIC
    @param {str}: wrfout_prefix: the prefix of WRF output files, default is "wrfout_d03_";if wrfout_dir contains different domains, the prefix should be specified for each domain.like "wrfout_d01_" for domain 1.
    @return: 
    {dict}: variables_data: the extracted variables data
    {dict}: variables_units: the units of extracted variables
    {dates}: the dates of extracted data
    示例：
    dict_variable_data,dict_variable_units=extract_wrf_hourly_data(wrfout_dir, output_wrf_file,variables=["Temperature_2m", "Humidity_2m", "Pressure_Surface", "Wind_Speed_10m", "Wind_Direction_10m"],output_wrf_file_format="NETCDF3_CLASSIC",wrfout_prefix="wrfout_d03_")
    '''
    # 遍历所有 WRF 输出文件，提取数据
    wrf_files = sorted(glob.glob(os.path.join(wrf_out_dir, f"{wrf_out_prefix}*")))
    # combine_netcdf_files_with_mfdataset(wrf_files, output_wrf_file+".mfdataset.nc", variables, output_file_format=output_wrf_file_format)
    dict_variables_data = {} 
    dict_variables_dims,global_attrs_template,dates,latitudes,longitudes = {},{},None,None,None   
    for wrf_file in tqdm(wrf_files, desc="读取WRF输出文件"):#[0:3]:
        dict_data,variables_units,dims,global_attrs_template,lats,lons,times=extract_data_from_wrf_file(wrf_file,variables=variables)# extract_wrf_data_from_file(wrf_file,variables=variables)
        if len(dict_variables_dims)==0:
            dict_variables_dims,global_attrs_template,dates,latitudes,longitudes=dims.copy(),global_attrs_template,times,lats,lons          
        for name, data in dict_data.items():
            if dict_variables_data.get(name) is None:
                dict_variables_data[name] = []
            if len(dict_variables_data[name])>0 and  data.ndim!=dict_variables_data[name][-1].ndim:
                print(f"{wrf_file} 变量 {name} 维度不一致，跳过")
                continue
            dict_variables_data[name].append(data)
    variables_data = {}
    # 将所有数据沿时间维度拼接
    for var, data in dict_variables_data.items():        
        variables_data[var] = np.concatenate(data, axis=0)  # 沿时间维度拼接         
    save_wrf_data(output_wrf_file,dict_variables_dims,global_attrs_template,variables_data,variables_units=variables_units,output_file_format=output_wrf_file_format)
    # dates=pd.date_range(start_time,periods=len(variables_data[0]),freq="H")
    return variables_data,variables_units,dates

    
def calculate_specified_level_metric(file,target_variable):
    '''
    TODO: 待验证
    @description: 计算指定高度的气象变量
    @param {str}: file: WRF 输出文件
    @param {str}: target_variable: 目标气象变量
    @return: {np.array}: 指定高度的气象变量    
    '''
    wrfin = Dataset(file)
    rh = getvar(wrfin, target_variable)
    height = getvar(wrfin, "height_agl")
    pblh = getvar(wrfin, "PBLH")
    rh_pblh = interplevel(rh, height, pblh)
    return rh_pblh

# 定义函数：计算相对湿度 (RH)
def calculate_relative_humidity(q2, psfc, t2):
    '''
    TODO: 待验证
    @description: 计算相对湿度 (RH)
    @param {np.array}: q2: 2m气压 (Pa)
    @param {np.array}: psfc: 地面气压 (Pa)
    @param {np.array}: t2: 2m温度 (K)
    '''
    # 常量定义
    epsilon = 0.622  # 水汽与干空气的比值
    e0 = 6.112  # hPa
    t2_celsius = t2 - 273.15  # 转换为摄氏度
    es = e0 * np.exp((17.67 * t2_celsius) / (t2_celsius + 243.5))  # 饱和水汽压 (hPa)
    e = (q2 * psfc) / (epsilon + (1 - epsilon) * q2)  # 水汽压 (hPa)
    rh = (e / es) * 100  # 相对湿度 (%)
    return rh

def calculate_rh2(q2, psfc, t2): 
    '''
    TODO: 待验证
    @description: 计算2m相对湿度 (RH2)；
    @param {np.array}: q2: 2m气压 (Pa)
    @param {np.array}: psfc: 地面气压 (Pa)
    @param {np.array}: t2: 2m温度 (K)
    '''
    rh2 = q2 / (0.6220 * (611 * np.exp(19.8644 - 5423 / t2) / (psfc - 611 * np.exp(19.8644 - 5423 / T2)))) * 100
    return rh2

# 定义函数：计算风速和风向
def calculate_wind_speed_direction(u10, v10):
    '''
    TODO: 待验证
    @description: 计算风速和风向
    @param {np.array}: u10: 10m风速 (m/s)
    @param {np.array}: v10: 10m风速 (m/s)
    @return: {np.array}: wind_speed: 风速 (m/s)
    @return: {np.array}: wind_direction: 风向 (度)
    '''
    wind_speed = np.sqrt(u10**2 + v10**2)  # 风速
    wind_direction = (270 - np.arctan2(v10, u10) * 180 / np.pi) % 360  # 风向
    return wind_speed, wind_direction   
        
def combine_wrf_files(wrf_files,output_file,variables=None,output_file_format="NETCDF3_CLASSIC"):
    '''
    TODO: 待验证
    @description: 合并多个 WRF 输出文件
    @param {list}: wrf_files: WRF 输出文件列表
    @param {str}: output_file: 输出文件名
    @param {list}: variables: 需要提取的变量列表
    @param {str}: output_file_format: 输出文件格式
    @return: None
    '''
    import xarray as xr
    try:
        # import xarray as xr
        # ds_input1 = xr.open_mfdataset(wrf_files, combine='Times')
         # 指定要合并的维度 
        # 使用 MFDataset 函数打开多个文件并指定要合并的维度 : 主要用于读取多个 NetCDF 文件的内容;它是只读的      
        f = nc.MFDataset(wrf_files) 
        dict_variable_data,dict_variable_units,dict_dims,global_attrs_template,lats,lons,time={},{},{},{},None,None,None 
        for variable_name in variables:
            wrf_var=wrf_variables[variable_name][0]
            dict_variable_data[variable_name] = f.variables[wrf_var][:]
            dict_variable_units[variable_name] = f.variables[wrf_var].units
            dict_dims[variable_name] = f.variables[wrf_var].dimensions            
            
        # 使用 xarray 打开并合并多个文件
        # ds = xr.open_mfdataset(wrf_files, parallel=True)
        # # 保存合并后的文件为新的 NetCDF 文件   
        # ds.to_netcdf(output_file+"_xarray.nc")
        # # 关闭数据集
        # ds.close()
        # # 打开多个文件并合并它们
        # datasets = [xr.open_dataset(file) for file in wrf_files]
        # # 合并数据集
        # combined_dataset = xr.concat(datasets, dim='Time')    
        # # 将合并后的数据集保存为一天 24 小时的 NetCDF 文件
        # combined_dataset.to_netcdf(output_file)    
        # # 关闭数据集
        # combined_dataset.close()        
        
    except Exception as e:
        print(f"Error combining WRF files: {e}")

def extract_wrf_data_within_start_end_time(wrf_file,start_time,end_time,variables=None):
    '''
    TODO: 待验证
    @description: 从 WRF 输出文件中提取指定时间段的数据
    @param {str}: wrf_file: WRF 输出文件
    @param {datetime}: start_time: 开始时间
    @param {datetime}: end_time: 结束时间
    '''
    from wrf import getvar, extract_times, ALL_TIMES
    nc_dataset = Dataset(wrf_file)
    times = extract_times(nc_dataset, timeidx=ALL_TIMES)  # 返回 datetime 对象列表peroid
    # 筛选指定时间段的数据
    time_mask = [(t >= start_time) & (t <= end_time) for t in times]
    # 提取变量（例如 2 米温度 T2）
    t2 = getvar(nc_dataset, "T2", timeidx=ALL_TIMES)  # 提取所有时间的 T2 数据
    filtered_t2 = t2[time_mask, :, :]  # 根据时间掩码筛选数据 
    return filtered_t2

if __name__ == '__main__':
    import os
    import glob
    import numpy as np
    import pandas as pd
    
  
    # 定义文件路径和变量
    wrf_out_dir = "/bigdata/wrf_output_file/liuziyi/PRDNEW"  # WRF 输出文件的目录
    output_wrf_file = "/DeepLearning/mnt/Devin/data/wrf_out/2022082600-2022100200_wrf_data.nc"  # 输出文件名
    variables=["Planetary_Boundary_Layer_Height", "Temperature_2m", "Humidity_2m", "Pressure_Surface", "Wind_Speed_10m", "Wind_Direction_10m","Solar_Radiation","Precipitation","Cloud_Cover"]
    # variables=["Temperature_2m", "Humidity_2m", "Wind_Speed_10m", "Wind_Direction_10m"]

    wrf_out_prefix="wrfout_d03_"
    output_wrf_file_format="NETCDF3_CLASSIC"
    wrf_files = sorted(glob.glob(os.path.join(wrf_out_dir, f"{wrf_out_prefix}*")))    
    extracted_wrf_variables =list(set([var for variable in variables for var in wrf_variables[variable]]) if variables is not None else [])
    # extracted_wrf_variables=["T2","PBLH",]#,'XLAT', 'XLONG'
    # combine_netcdf_files_with_mfdataset(wrf_files, output_wrf_file+".mfdataset3.nc", extract_wrf_variables=extracted_wrf_variables, output_file_format=output_wrf_file_format, support_wrf_python=True)
    wrf_file=output_wrf_file+".mfdataset3.nc"
    x,y,lons,lats,dx,dy,cart_proj=get_wrf_info(wrf_file,variable_name=extracted_wrf_variables[0])
    date_peroid=pd.date_range("2022-09-01 00:00:00", "2022-10-01 00:00:00", freq="H")
    #方法1：直接提取
    dict_variable_data,dict_variable_units,dict_dims,global_attrs_template,lats,lons,times=extract_data_from_wrf_file(wrf_file,variables=variables,date_peroid=date_peroid,convert_to_beijing_time=True)
    #方法2：提取指定时间段的数据
    #region 方法1：使用 extract_data_from_wrf_file和extract_wrf_hourly_data_without_combining_files 函数提取数据
    # extract_wrf_data_from_file(wrf_file,variables=variables)
    # dates=pd.date_range(start_time,periods=len(dict_variable_data.pop(0)),freq="H")    
    variables_data,variables_units,start_time=extract_wrf_hourly_data_without_combining_files(wrf_out_dir, output_wrf_file,variables=variables,wrf_out_prefix="wrfout_d03_",output_wrf_file_format="NETCDF3_CLASSIC")
    #endregion 方法1：使用 extract_data_from_wrf_file和extract_wrf_hourly_data_without_combining_files 函数提取数据
    output_wrf_file = f"/DeepLearning/mnt/Devin/data/wrf_out/{date_peroid[0].strftime('%Y%m%d%H')}-{date_peroid[-1].strftime('%Y%m%d%H')}_wrf_data.nc"  # 输出文件名
    variables_data,variables_units,start_time=extract_wrf_hourly_data(wrf_out_dir, output_wrf_file,variables=variables,wrf_out_prefix="wrfout_d03_",output_wrf_file_format="NETCDF3_CLASSIC",date_peroid=date_peroid,convert_to_beijing_time=True)
    print(f"数据已保存到 {output_wrf_file}")



"""
Author: Devin
Date: 2023-12-23 18:18:13
LastEditors: Devin
LastEditTime: 2024-11-28 16:19:09
FilePath: \PythonDemo\esil\RSM\Model_Property.py
Description: 

Copyright (c) 2024 by Devin, All Rights Reserved. 
"""

import netCDF4 as nc
from pyproj import pyproj, transform
import numpy as np


class model_attribute:

    def __init__(self, nc_file):
        nc_data = nc.Dataset(nc_file)
        center_lon = nc_data.getncattr("XCENT")
        center_lat = nc_data.getncattr("YCENT")
        lat_1 = nc_data.getncattr("P_ALP")
        lat_2 = nc_data.getncattr("P_BET")
        proj4_string = ""
        # Check conditions and set projection string accordingly
        if center_lat == 40 and center_lon == -97:
            proj4_string = f"+proj=lcc +lat_1={lat_1} +lat_2={lat_2} +lat_0={center_lat} +lon_0={center_lon} +a=6370000.0 +b=6370000.0"
        else:
            proj4_string = f"+x_0=0 +y_0=0 +lat_0={center_lat} +lon_0={center_lon} +lat_1={lat_1} +lat_2={lat_2} +proj=lcc +ellps=WGS84 +no_defs"
        self.proj4_string = proj4_string
        projection = pyproj.Proj(proj4_string)
        self.x_orig = nc_data.getncattr("XORIG")
        self.y_orig = nc_data.getncattr("YORIG")
        self.x_resolution = nc_data.getncattr("XCELL")
        self.y_resolution = nc_data.getncattr("YCELL")
        self.cols = nc_data.getncattr("NCOLS")
        self.rows = nc_data.getncattr("NROWS")
        self.min_x = self.x_orig + self.x_resolution / 2
        self.min_y = self.y_orig + self.y_resolution / 2
        self.max_x = self.min_x + self.cols * self.x_resolution
        self.max_y = self.min_y + self.rows * self.y_resolution
        self.lon_start, self.lat_start = projection(
            self.min_x, self.min_y, inverse=True
        )
        self.lon_end, self.lat_end = projection(self.max_x, self.max_y, inverse=True)
        import cartopy.crs as ccrs

        self.projection = ccrs.LambertConformal(
            central_longitude=center_lon,
            central_latitude=center_lat,
            standard_parallels=(lat_1, lat_2),
        )

        self.is_BC = "PERIM" in nc_data.dimensions
        if self.is_BC:
            self.rows = self.rows + 2  # 加上下外边界网格
            self.cols = self.cols + 2  # 加前后外边界网格
            self.start_date = nc_data.getncattr("SDATE")
            if self.start_date:
                from esil import date_helper

                self.start_date = date_helper.convert_julian_regular_date(
                    self.start_date
                )
            self.min_x = float(self.x_orig - self.x_resolution / 2.0)  # 扩大一个网格
            tmp = (self.cols + 1) * self.x_resolution
            # 取得最大坐标
            self.max_x = float(tmp + self.min_x)
            # 取得最小y坐标
            self.min_y = float(self.y_orig + self.y_resolution / 2.0)
            # 1~112行之间共有111段，所以1~112之间的总长度是：（112-1）*网格高度
            tmp = (self.rows + 1) * self.y_resolution
            # 取得最大坐标
            self.max_y = float(tmp + self.min_y)

            self.x_coords_bc, self.y_coords_bc, self.x_coords, self.y_coords = (
                self.get_xy_coords()
            )

            # self.lons, self.lats = projection(self.x_coords, self.y_coords , inverse=True)
            # 将 Lambert 投影坐标系下的 x、y 转换为经纬度坐标系下的 x、y
            self.lats, self.lons = transform(
                projection, "EPSG:4326", self.x_coords_bc, self.y_coords_bc
            )  # 'EPSG:4326' 表示经纬度坐标系
            # lon, lat = projection(self.x_coords[0], self.y_coords[0], inverse=True)
            # print( lon, lat )
        else:
            self.x_coords, self.y_coords = self.get_xy_coords()
            grid_x, grid_y = np.meshgrid(self.x_coords, self.y_coords)
            self.lats, self.lons = transform(projection, "EPSG:4326", grid_x, grid_y)
            # self.lons, self.lats = projection(self.x_coords, self.y_coords, inverse=True)

    def get_xy_coords(self):
        try:
            if self.is_BC:
                cols, rows = self.cols, self.rows
                lon_values, lat_values = [], []
                for col in range(cols):
                    lon_values.append(self.min_x + col * self.x_resolution)
                for row in range(rows):
                    lat_values.append(self.min_y + row * self.y_resolution)
                float_x_coords, float_y_coords = [], []
                # 从左下角第一个网格开始，逆时针获取每个网格的经纬度
                for col in range(cols):
                    float_y_coords.append(lat_values[0])
                    float_x_coords.append(lon_values[col])
                for row in range(1, rows - 1):
                    float_y_coords.append(lat_values[row])
                    float_x_coords.append(lon_values[cols - 1])
                for col in range(cols - 1, -1, -1):
                    float_y_coords.append(lat_values[rows - 1])
                    float_x_coords.append(lon_values[col])
                for row in range(rows - 2, 0, -1):
                    float_y_coords.append(lat_values[row])
                    float_x_coords.append(lon_values[0])
                return float_x_coords, float_y_coords, lon_values, lat_values
            else:
                row_count = self.rows
                col_count = self.cols
                float_x_coords = [0.0] * col_count
                float_y_coords = [0.0] * row_count
                for col in range(col_count):
                    float_x_coords[col] = self.min_x + col * self.x_resolution
                for row in range(row_count):
                    float_y_coords[row] = self.min_y + row * self.y_resolution
        except Exception as ex:
            print(f"Error: {ex}")
        return float_x_coords, float_y_coords

    def get_xy_coordinates(self, show_lonlat=False):
        """
        description: 获取需要显示的x、y坐标
        param {class} model，模型属性对象
        param {bool} show_lonlat，默认为False，如果为True，则返回经纬度坐标
        return {numpy.ndarray(2D), numpy.ndarray(2D)}
        """
        if show_lonlat:
            x, y = self.lons, self.lats
        else:
            x = np.linspace(1, self.cols, self.cols)
            y = np.linspace(1, self.rows, self.rows)
            x, y = np.meshgrid(x, y)
        return x, y



import numpy as np


def inverse_distance_weighting(x, y, z, xi, yi, power=2):
    """
    @description:反距离平方 IDW 插值计算
    x,y为原数组一维长度
    z的长度是len(x)*len(y) 二维数组
    xi 为插值后二维数组，即meshgrid后
    yi 为插值后二维数组，即meshgrid后
    """

    from scipy.spatial.distance import cdist

    xx, yy = np.meshgrid(x, y)
    xy = np.column_stack([xx.flatten(), yy.flatten()])
    # xi, yi = np.meshgrid(xi, yi)
    xyi = np.column_stack([xi.flatten(), yi.flatten()])

    distances = cdist(xy, xyi)
    weights = 1 / (distances + 1e-10) ** power  # 避免除以零，加上一个很小的值
    weights /= np.sum(weights, axis=0)  # 在列方向上归一化权重

    # 将 z 展开成一维数组，方便后续计算
    z_flat = z.flatten()

    # 通过对每个点的权重进行加权求和得到插值结果
    zi = np.dot(weights.T, z_flat)
    zi = zi.reshape(xi.shape)
    return zi


def interpolate(x, y, data, interpolation_number, method="griddata-cubic"):
    """
    @description:对数据进行插值，返回插值后x,y坐标数组及插值结果

    Args:
        x (1D array): x坐标一维数组
        y (1D array): y坐标一维数组
        data (2D array):需要插值的数据，二维数组,长度x*y，其中x,y需要通过np.meshgrid方法将其变为二维数组，然后再用flatten将其转为一维数组，此时x和y的长度变为原始x*y的长度
            lon, lat = np.meshgrid(x, y)
            x = lon.flatten()
            y = lat.flatten()
        interpolation_number:插值个数，大于x和y的最大长度，否则取x和y中的最大长度
        method:插值方法;
                1."griddata-cubic":三次样条插值。使用三次样条函数在数据点之间进行插值。，默认采用此方法
                2."griddata-linear":线性插值。在三角网格中使用线性插值法计算点的值。
                3."griddata-cubic":最近邻插值。根据最近邻点的值来估计目标点的值。
                4."kriging-linear": variogram_models.linear_variogram_model,
                5."kriging-power": variogram_models.power_variogram_model,
                6."kriging-gaussian": variogram_models.gaussian_variogram_model,
                7."kriging-spherical": variogram_models.spherical_variogram_model,
                8."kriging-exponential": variogram_models.exponential_variogram_model,
                9."kriging-hole-effect": variogram_models.hole_effect_variogram_model,
                10."idw-power"：反距离权重方法，power是具体方次，默认是反距离平方，即idw-2；它使用距离的平方的倒数作为权重来估计未知位置的值。这种方法中，较远的点权重较小，而较近的点权重较大
    Returns:
        grid_x: 插值后x坐标:list of ndarrays
        grid_y:插值后y坐标:list of ndarrays
        grid_data:插值后data:list of ndarrays
    """

    interpolated_result = []
    ori_grid_x, ori_grid_y = np.meshgrid(x, y)  # 对原始数据中lon,lat两个维度进行网格化
    # 获取插值个数
    num = max(len(x), len(y))
    if interpolation_number < num:
        interpolation_number = num
    # num = num + interpolation_number
    interpolated_x = np.linspace(x.min(), x.max(), interpolation_number)
    interpolated_y = np.linspace(y.min(), y.max(), interpolation_number)
    grid_x, grid_y = np.meshgrid(interpolated_x, interpolated_y)

    if "griddata" in method:
        from scipy.interpolate import griddata

        sub_method = method.split("-")[-1]  # 取最后一个
        interpolated_result = griddata(
            (ori_grid_x.flatten(), ori_grid_y.flatten()),
            data.flatten(),
            (grid_x, grid_y),
            method=sub_method,
        )
    elif "kriging" in method:
        from pykrige.ok import OrdinaryKriging

        sub_method = method.split("-")[-1]  # 取最后一个
        # 创建一个克里金插值对象
        ok = OrdinaryKriging(
            ori_grid_x.flatten(),
            ori_grid_y.flatten(),
            data.flatten(),
            variogram_model=sub_method,
            verbose=False,
            enable_plotting=False,
        )
        interpolated_result, ss = ok.execute("grid", interpolated_x, interpolated_y)
    elif "idw" in method:
        power = (int)(method.split("-")[-1])  # 取最后一个
        # 进行反距离平方 IDW 插值
        interpolated_result = inverse_distance_weighting(
            x, y, data, grid_x, grid_y, power=power
        )
    else:
        print(f"Failed to found this method:{method}")
    return (grid_x, grid_y, interpolated_result)


def smooth_processing_data(data, method="gaussian", sigma=1, size=3):
    """
    对数据进行平滑处理

    Args:
        data (2D array):需要平滑处理的数据，二维数组
        method:平滑方法;
                1."gaussian":高斯滤波
                2."uniform":均值滤波
                3."median":中值滤波
        sigma:高斯滤波的标准差，用于控制平滑程度;sigma 值越大，平滑效果越明显，但也可能导致数据的细节丢失
        小 sigma 值：适用于需要保留更多细节的情况。常见取值范围是 0.5 到 1.5。
        中等 sigma 值：适用于需要适度平滑的情况。常见取值范围是 1.5 到 3。
        大 sigma 值：适用于需要强烈平滑的情况。常见取值范围是 3 到 5 或更大
        size:均值滤波和中值滤波的窗口大小，用于控制平滑程度;窗口大小越大，平滑效果越明显，但也可能导致数据的细节丢失
    调用示例：smooth_processing_data(data,method='gaussian',sigma=1,size=3)
    Returns:
        smoothed_data:平滑后的数据
    """
    if "gaussian" in method:
        from scipy.ndimage import gaussian_filter

        smoothed_data = gaussian_filter(data, sigma=sigma)
    elif "uniform" in method:
        from scipy.ndimage import uniform_filter

        smoothed_data = uniform_filter(data, size=size)
    elif "median" in method:
        from scipy.ndimage import median_filter

        smoothed_data = median_filter(data, size=size)
    else:
        print(f"Failed to found this method:{method}")
    return smoothed_data


if __name__ == "__main__":
    from map_helper import show_single_map
    from netcdf_helper import get_extent_data

    nc_file = r"D:\Project\G广东省主要温室气体排放反演与大气污染精准溯源服务项目\090数据\ENKF 全球反演结果\NJU.GLOBAL.FLUX.1x1.202201.nc"
    variable_name = "bio_prio"
    extent = [70.25, 136.25, 4.75, 55.25]
    data, unit = get_extent_data(nc_file, variable_name)
    # data=get_extent_data(nc_file,variable_name,extent=extent)
    x = data.lon.values
    y = data.lat.values
    data = data.values
    grid_x, grid_y = np.meshgrid(x, y)

    # data[:] = np.nan
    show_single_map(
        grid_x,
        grid_y,
        grid_concentration=data,
        title=variable_name + " origin data",
        cmap="gray",
    )

    # grid_x,grid_y,interpolated_result=interpolate(x,y,data,100)
    # plotmap(grid_x,grid_y,grid_concentration=interpolated_result,title=variable_name+ " with griddata-cubic")
    # #grid_x, grid_y, interpolated_result = interpolate(x=x, y=y, data=data,interpolation_number= 100,method='griddata-linear')
    # grid_x, grid_y, interpolated_result = interpolate(x=x, y=y, data=data,interpolation_number= 100,method='kriging-gaussian')
    # plotmap(grid_x, grid_y, grid_concentration=interpolated_result, title=variable_name + " with kriging-gaussian")
    # grid_x, grid_y, interpolated_result = interpolate(x=x, y=y, data=data, interpolation_number=100, method='idw-2')
    # plotmap(grid_x, grid_y, grid_concentration=interpolated_result, title=variable_name+" with idw-2")
    print("success")



import os.path
import numpy as np
import pandas as pd
import esil.log_helper as log
import sys
import fnmatch
import random

class rsm_helper:

    def create_grid_from_receptor_grid_file(self,file_path, output_path='grid.npy'):
        '''
        创建DeepRSM所需的grid.npy文件
        :param file_path: 基于Receptor grid file（RSM-VAT 中的网格受体文件）进行获取，对同个网格而言，只保留网格占比最大的记录
        :param output_path: 输出文件路径，默认在当前目录下的grid.npy,用户可以自行指定路径
        :return: 没有返回值
        '''
        # 读取文件
        max_ratios = {}
        with open(file_path, 'r') as file:
            for line in file:
                values = line.strip().split()
                id_val = int(values[0])  # ID
                col = int(values[1])  # Col
                row = int(values[2])  # Row
                ratio = float(values[3])  # Ratio
                current_max_ratio, current_id = max_ratios.get((col, row), (float('-inf'), None))
                if ratio > current_max_ratio:
                    max_ratios[(col, row)] = (ratio, id_val)

        # 输出结果
        print("ID Col Row Ratio Code City")
        # 获取 Col 和 Row 的最大 Ratio 记录的 ID，并存储到 NumPy 二维数组中
        max_col = max(col for col, _ in max_ratios.keys())
        max_row = max(row for _, row in max_ratios.keys())
        result_array = np.zeros((max_row, max_col), dtype=int)  # 为了和后面deeprsm脚本保持一致，这里改为先行后列
        for (col, row), (_, id_val) in max_ratios.items():
            result_array[row - 1][col - 1] = id_val
            # 保存为 .npy 文件
        np.save(output_path, result_array)

    #@staticmethod # Python 没有严格的静态类的概念，但可以通过在类中使用 @staticmethod 装饰器来模拟类级别的静态方法
    def normalize_pollutant(self,pollutant):
        '''
        规范化污染物书写，如NOX统一规范为NOx,PM25改为PM2.5
        :param pollutant:待处理的物种名称
        :return:
        '''
        result = ''
        pollutant = pollutant.upper()
        if pollutant == 'NOX':
            result = 'NOx'
        elif pollutant == 'SO2':
            result = 'SO2'
        elif pollutant == 'NH3':
            result = 'NH3'
        elif pollutant in ('VOCS', 'VOC'):
            result = 'VOC'
        elif pollutant == 'POA':
            result = 'POA'
        elif pollutant in ('PM2.5', 'PM', 'PM25'):
            result = 'PM2.5'
        else:
            result = pollutant
        return result

    #@staticmethod # Python 没有严格的静态类的概念，但可以通过在类中使用 @staticmethod 装饰器来模拟类级别的静态方法
    def get_mapping_data(self,emission_file,cmaq_name='ACONC'):
        '''
        根据RSM-VAT中 Emission matrix文件，获取每个case所对应的名称
        如{'BaseCase': 'ACONC.1', 'Ctrl_All': 'ACONC.21', 'Ctrl_A': 'ACONC.22', 'Ctrl_B': 'ACONC.23', 'Ctrl_C': 'ACONC.24', 'Ctrl_D': 'ACONC.25', 'Ctrl_E': 'ACONC.26', 'Ctrl_F': 'ACONC.27'}
        :param emission_file: Emission matrix文件
        :param cmaq_name: 用做RSM-VAT的模型结果文件名称，如ACONC
        :return: 返回字典，如{'BaseCase': 'ACONC.1', 'Ctrl_All': 'ACONC.21', 'Ctrl_A': 'ACONC.22', 'Ctrl_B': 'ACONC.23', 'Ctrl_C': 'ACONC.24', 'Ctrl_D': 'ACONC.25', 'Ctrl_E': 'ACONC.26', 'Ctrl_F': 'ACONC.27'}
        '''
        dic_mapping_data={}
        try:
            dt = pd.read_csv(emission_file)
            for rowIndex in range(len(dt.index)):
                columnIndex = 0
                factor_infos = []
                for column in dt.columns:
                    arr = column.split(')').pop().replace('\\', '/').replace('_', '/').split('/')
                    if len(arr) >= 3:
                        factor = {
                            "Region": arr[0],
                            "Pollutant": self.normalize_pollutant(arr[1]),
                            "Source": arr[2],
                            "BaseEmissionRatio": float(dt.iloc[rowIndex,columnIndex]),
                            "Index": columnIndex
                        }
                        factor_infos.append(factor)
                    columnIndex += 1
                case_num = f"{cmaq_name}.{dt.iloc[rowIndex,0]}"  # Defaulting to the first column as scenario number
                temp = [f for f in factor_infos if f['Pollutant'] != 'PM2.5']
                total = sum(m['BaseEmissionRatio'] for m in temp)
                if total == len(temp):  # Baseline
                    is_all_base_case = len([p for p in factor_infos if p['BaseEmissionRatio'] == 1]) == len(factor_infos)
                    if is_all_base_case:
                        dic_mapping_data["BaseCase"] = case_num
                elif total == 0:  # Shutdown in all regions
                    dic_mapping_data["Ctrl_All"] = case_num
                elif total == len(temp) - 4:  # Shutdown in a single region
                    # 使用字典来模拟LINQ中的GroupBy和Sum操作
                    grouped = {}
                    for item in temp:
                        region = item['Region']
                        ratio = item['BaseEmissionRatio']
                        if region not in grouped:
                            grouped[region] = {'Sum': 0}
                        grouped[region]['Sum'] += ratio
                    # 找到和为0的第一个Region
                    region = next((key for key, value in grouped.items() if value['Sum'] == 0), None)
                    if region and f"Ctrl_{region}" not in dic_mapping_data:
                            dic_mapping_data[f"Ctrl_{region}"] = case_num

        except Exception as ex:
            log.write(ex,sys._getframe())
        return (dic_mapping_data,factor_infos)

    def get_species_from_mapping_file(self,path):
        '''
        获取18种species（"HNO3", "H2O2", "N2O5", "NO2", "FORM", "HONO", "NO", "O3", "SO2", "NH3", "OH", "ISOP", "TERP", "ASO4IJ", "ANO3IJ", "ANH4IJ", "ASOCIJ", "PM25" ）在cmaq文件中对应的物种名称
        :param path:RSM-VAT中的Species Mapping file
        :return: 返回一一对应的物种字典
        '''
        try:
            if not os.path.exists(path):
                return None
            dic_species = {}  # 映射文件中的物种信息
            dt_species = pd.read_csv(path)
            species_num = len(dt_species.index)
            for i in range(species_num):  # 获取映射文件中的物种映射关系
                dic_species[dt_species.loc[i, "Species Name"]] = dt_species.loc[i, "Species Name In Model File"]
        except Exception as ex:
            log.write(ex, sys._getframe())
        return dic_species

    def reset_to_default_path(self,savepath):
        '''
        将经过处理的文件路径恢复回默认
        :return:
        '''
        path = savepath.replace("*", " ")
        return path

    def GetYear(self,date_str):
        '''
        从netcdf中的SDATE变量中提取模型数据年份
        :param date_str:
        :return:
        '''
        dateTime = str(date_str)
        year = ''
        if len(dateTime) >= 4:
            year = dateTime[:4]
        return year

    # 生成指定范围内的随机数组（允许重复值）
    def generate_random_array_with_duplicates(self, start, end, size):
        random_array = [random.uniform(start, end) for _ in range(size)]
        return random_array


    def combineNPY(self,inputFilePath, outputFileName='', filter='*data_species_zero.npy', isRemoveFile=False):
        '''
         多个npy文件合并成一个npy文件
         读取文件路径下的满足filter特征的文件，并进行合并生成文件名为key，value为data的字典
        :param inputFilePath:
        :param outputFileName:
        :param filter:
        :param isRemoveFile:
        :return:
        '''
        try:
            dic = {}
            outputDir = ''
            if inputFilePath.find(':') < 0:  # ':' in inputFilePath
                inputFilePath = os.path.abspath(inputFilePath)
            for root, dirs, files in os.walk(inputFilePath):
                outputDir = root
                for filename in fnmatch.filter(files, filter):
                    real_path = (os.path.join(root, filename))
                    real_data = np.load(real_path, allow_pickle=True)  # 类型是numpy array
                    dic[filename] = real_data
                    if isRemoveFile:
                        os.remove(real_path)
                    if len(outputFileName) == 0:
                        outputFileName = filename.replace('.npy', '_dict.npy')
                        '''
                        if filter.find('zero')>=0:
                            outputFileName = filename.replace('.npy' , 'dict')
                        #elif filter.find('base')>=0:
                        else:
                            outputFileName = filename.replace('.npy' , '_dict.npy')
                        '''
            np.save(os.path.join(outputDir, outputFileName), dic)
        except Exception as e:
            log.write(e, sys._getframe())

    def sperateNPY(self,inputFilePath, isRemoveFile=False):
        '''
        将字典中数据安装key值文件名进行输出
        :param inputFilePath:
        :param isRemoveFile:
        :return:
        '''
        try:
            data_dic = np.load(inputFilePath, allow_pickle=True).item()  # 字段读取
            dir = os.path.dirname(inputFilePath)  # 获取路径字符串中的目录
            # file=os.path.basename(path)#获取路径字符串中的文件名
            for key in data_dic:
                filepath = os.path.join(dir, key)
                np.save(filepath, data_dic[key])
            if isRemoveFile == True:
                os.remove(inputFilePath)
        except Exception as e:
            log.write(e, sys._getframe())

    def read_oos_file(self,oos_emission_file,cmaq_name='ACONC'):
        '''
                根据RSM-VAT中 oos emission文件，获取每个case所对应的矩阵
                如{'BaseCase': 'ACONC.1', 'Ctrl_All': 'ACONC.21', 'Ctrl_A': 'ACONC.22', 'Ctrl_B': 'ACONC.23', 'Ctrl_C': 'ACONC.24', 'Ctrl_D': 'ACONC.25', 'Ctrl_E': 'ACONC.26', 'Ctrl_F': 'ACONC.27'}
                :param oos_file: oos文件
                :param cmaq_name: 用做RSM-VAT的模型结果文件名称，如ACONC
                :return: 返回tuple, item1=字典（key=case_name,value=emission_ctrl),item2=factors集合
                '''
        dic_case_emission = {}
        try:
            dt = pd.read_csv(oos_emission_file)
            for rowIndex in range(len(dt.index)):
                case_num = f"{cmaq_name}.{dt.iloc[rowIndex, 0]}"
                emission_ctrl=dt.iloc[rowIndex, 1:].values.astype(float)
                dic_case_emission[case_num]=emission_ctrl
            columnIndex = 0
            factor_infos = []
            for column in dt.columns:
                arr = column.split(')').pop().replace('\\', '/').replace('_', '/').split('/')
                if len(arr) >= 3:
                    factor = {
                        "Region": arr[0],
                        "Pollutant": self.normalize_pollutant(arr[1]),
                        "Source": arr[2],
                        "BaseEmissionRatio": 1,
                        "Index": columnIndex
                    }
                    factor_infos.append(factor)
                columnIndex += 1
        except Exception as ex:
            log.write(ex, sys._getframe())
        return (dic_case_emission, factor_infos)

if __name__=="__main__":
    helper=rsm_helper()
    dict,factors=helper.get_mapping_data(r'E:\paper_data\PRD_Daily_DeepRSM\Config Files\Matrix_PRD.csv')
    dict1 = helper.get_species_from_mapping_file(r'E:\paper_data\PRD_Daily_DeepRSM\Config Files\SpeciesMapping.csv')
    print(dict)


'''
Author: Devin
Date: 2024-06-18 23:40:30
LastEditors: Devin
LastEditTime: 2024-08-22 17:20:07
Description: 

Copyright (c) 2024 by Devin, All Rights Reserved. 
'''
class FactorInfo:
    # def __init__(self):
    #     self.Region = ""
    #     self.Pollutant = ""
    #     self.Source = ""
    #     self.Min = 0.0
    #     self.Max = 1.2
    #     self.Limit = 0.2
    #     self.BaseEmissionRatio = 0.0
    #     self.Index = 0
    def __init__(self, region, pollutant, source, value, index=None, min=None, max=None):
        self.Region = region
        self.Pollutant = pollutant
        self.Source = source
        self.Min = min if min is not None else 0.0
        self.Max = max if max is not None else 1.2      
        self.Value = value
        self.Index = index
       


import pyproj
import netCDF4 as nc


def get_proj4string(cmaq_file):
    '''
    根据CMAQ模拟结果获取该文件的投影信息
    :param cmaq_file:
    :return:
    '''
    nc_data = nc.Dataset(cmaq_file)
    center_lon = nc_data.getncattr('XCENT')
    center_lat = nc_data.getncattr('YCENT')
    lat_1= nc_data.getncattr('P_ALP')
    lat_2 = nc_data.getncattr('P_BET')
    proj4_string=''
    # Check conditions and set projection string accordingly
    if center_lat == 40 and center_lon == -97:
        proj4_string = f"+proj=lcc +lat_1={lat_1} +lat_2={lat_2} +lat_0={center_lat} +lon_0={center_lon} +a=6370000.0 +b=6370000.0"
    else:
        proj4_string = f"+x_0=0 +y_0=0 +lat_0={center_lat} +lon_0={center_lon} +lat_1={lat_1} +lat_2={lat_2} +proj=lcc +datum=wgs84 +no_defs"
    return proj4_string


def get_domain_corners_latlon(cmaq_file):
    '''
    根据CMAQ模拟结果获取模型域左下角和右上角的经纬度
    :param cmaq_file:CMAQ模拟结果文件路径
    :return:lon_start,lat_start,lon_end, lat_end
    '''
    nc_data = nc.Dataset(cmaq_file)
    center_lon = nc_data.getncattr('XCENT')
    center_lat = nc_data.getncattr('YCENT')
    lat_1= nc_data.getncattr('P_ALP')
    lat_2 = nc_data.getncattr('P_BET')
    proj4_string=''
    # Check conditions and set projection string accordingly
    if center_lat == 40 and center_lon == -97:
        proj4_string = f"+proj=lcc +lat_1={lat_1} +lat_2={lat_2} +lat_0={center_lat} +lon_0={center_lon} +a=6370000.0 +b=6370000.0"
    else:
        proj4_string = f"+x_0=0 +y_0=0 +lat_0={center_lat} +lon_0={center_lon} +lat_1={lat_1} +lat_2={lat_2} +proj=lcc +ellps=WGS84 +no_defs"
    projection=pyproj.Proj(proj4_string)
    x_orig= nc_data.getncattr('XORIG')
    y_orig = nc_data.getncattr('YORIG')

    x_resolution=nc_data.getncattr('XCELL')
    y_resolution = nc_data.getncattr('YCELL')
    cols= nc_data.getncattr('NCOLS')
    rows = nc_data.getncattr('NROWS')
    x_orig=x_orig+x_resolution/2
    y_orig = y_orig + y_resolution / 2
    x_end=x_orig+cols*x_resolution
    y_end=y_orig+rows*y_resolution
    lon_start,lat_start=projection(x_orig,y_orig, inverse=True)
    lon_end, lat_end = projection(x_end, y_end, inverse=True)
    return (lon_start,lat_start,lon_end, lat_end)



def transform_point(origin_x,origin_y,source_projection,target_projection):
    #投影坐标转换为经纬度坐标
    lon,lat=source_projection(origin_x,origin_y, inverse=True)
    #经纬度坐标转换为第二个投影坐标
    out_x,out_y=target_projection(lon,lat)
    return (out_x,out_y)
def transform_point_2_latlon(origin_x,origin_y,source_projection):
    # 投影坐标转换为经纬度坐标
    lon, lat = source_projection(origin_x, origin_y, inverse=True)
    return (lon, lat)
def transform_point_from_latlon(lon, lat,target_projection):
    # 经纬度坐标转换为第二个投影坐标
    out_x, out_y = target_projection(lon, lat)
    return (out_x, out_y)

def convert_cartopy_projection_2_pyproj(cartopy_projection):
    # 获取该投影对象的proj4字符串
    proj4_string = cartopy_projection.proj4_init
    # 创建对应的pyproj投影对象
    pyproj_projection = pyproj.Proj(proj4_string)
    return pyproj_projection

if __name__=="__main__":
    nc_file= r'/NetcdfReader/Paper preparation scripts/Concentration distribution/cmaq_data/ACONC_2019.1'
    proj4_string=get_proj4string(nc_file)
    lon_start,lat_start,lon_end, lat_end=get_domain_corners_latlon(nc_file)
    print(lon_start,lat_start,lon_end, lat_end)
    # 定义 proj4 字符串
    #proj4_string = '+proj=lcc +lat_1=33 +lat_2=45 +lat_0=40 +lon_0=-97 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs'

    # 测试投影转换
    # 例如，转换一个经纬度坐标为投影坐标
   #longitude = -95.5  # 示例经度
    #latitude = 40.2  # 示例纬度

    # 将经纬度坐标转换为投影坐标
    #x, y = projection(longitude, latitude)


from setuptools import setup, find_packages

setup(
    name="esil",
    version="1.3.5",
    author="Devin Long",
    author_email="long.sc@qq.com",
    description="commonly used functions written by Devin Long",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Devin-Long-7/esil",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "numpy",
        "pandas",
        "xarray",
        "scipy",
        "tqdm",
        "netCDF4",
        "pyproj",
        "SQLAlchemy",
        "chardet",
        "sympy",
        "shapely",
        "pytz",
        # "pykrige",
        # "matplotlib<3.9,>=1.5",  # Basemap 要求的 Matplotlib 版本为小于3.9且大于等于1.5
        "cartopy",
        "geopandas",
    ],
)
